/**********************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
Developer: Robert Tupelo-Schneck <schneck@cnri.reston.va.us>
Developer: Tommy Tracy <tjt7a@virginia.edu>
Developer: Michael Booth <mjb3sb@virginia.edu>
\**********************************************************************/

package net.cnri.apps.doserver;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.InetAddress;
import java.security.cert.Certificate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import net.cnri.apps.doserver.web.*;
import net.cnri.do_api.DataElement;
import net.cnri.do_api.DigitalObject;
import net.cnri.do_api.Repository;
import net.cnri.dobj.DOAuthentication;
import net.cnri.dobj.DOClient;
import net.cnri.dobj.DOConnection;
import net.cnri.dobj.DOConstants;
import net.cnri.dobj.DOException;
import net.cnri.dobj.HeaderItem;
import net.cnri.dobj.HeaderSet;
import net.cnri.dobj.PKAuthentication;
import net.cnri.dobj.SecretKeyAuthentication;
import net.cnri.dobj.StorageProxy;
import net.cnri.util.StringUtils;
import net.handle.hdllib.Util;

import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.eclipse.jetty.servlet.ServletHolder;
import org.verisign.joid.consumer.AuthenticationResult;
import org.verisign.joid.consumer.JoidConsumer;
import org.verisign.joid.consumer.OpenIdFilter;

/** HTTPConnectionHandler objects run in their own thread and are responsible for
 *  managing a single connection to the server.  Each connection may have
 *  numerous operation invocations which is why each connection handler has
 *  its own RequestHandlerPool to keep track of threads for operation invocation
 *  requests.
 */

class HTTPConnectionHandler
extends HttpServlet
{

    public static final boolean DEBUG = true;
    public static final String DEFAULT_MIME_TYPE = "application/octet-stream";
    public static final String DEFAULT_GET_ELEMENT = "content";
    public static final String DEFAULT_ELEMENT_REF_ATT = "content_element";
    public static final String CONTENT_DISPOSITION_HEADER = "Content-Disposition";
    public static final int RC_OK = 200;
    public static final String RC_STR_OK = "OK";
    public static final int RC_NOTFOUND = 404;
    public static final String RC_STR_NOTFOUND = "Not Found";
    private static final int NONCE_RANDOM_SIZE = 20;
    public static final String XML = "application/xml";
    public static final String HTML = "text/html";
    private static final String PLAINTEXT = "text/plain";
    private static final String DO_AUTHENTICATION_REQ_ATTRIBUTE = "net.cnri.apps.doserver.do_authentication";
    private static final String AUTHORIZATION_ATTRIBUTE = "net.cnri.apps.doserver.authorization";
    
    public static final String STATIC_URI_PREFIX = "/static/";
    
    public static final String OBJECT_ID_PARAM = "id";
    public static final String OPENID_ATTRIBUTE = "caller";
    public static final String OPENID_IDENTITY = "openid.identity";
    public static final int BUFFER_SIZE = 1024;

    private Main serverMain;
    private DOClient anonymousDOClient;
    private JoidConsumer joid = new JoidConsumer();
    private File staticBase;
    
    // Constructor; set main, construct headers and parameters maps
    HTTPConnectionHandler(Main main) {
      serverMain = main;
      anonymousDOClient = new DOClient(PKAuthentication.getAnonymousAuth());
      staticBase = new File(serverMain.getBaseFolder(), "html"+File.separator+"static"+File.separator);
    }


    /**
     * This should be called before starting the HTTP server in order to configure the 
     * servlet parameters.
     */
    void setupServletHolder(ServletHolder servletHolder) {
        servletHolder.setInitParameter("storeClassName", "org.verisign.joid.server.MemoryStore");
        servletHolder.setInitParameter("userManagerClassName", "org.verisign.joid.server.MemoryUserManager");
        try {
            Velocity.setProperty("resource.loader","file,class");
            Velocity.setProperty("class.resource.loader.class","org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader");
            Velocity.setProperty("file.resource.loader.path", new File(serverMain.getBaseFolder(), "html/templates").getAbsolutePath());
            Velocity.init();
        } catch (Exception e) {
            throw new AssertionError(e);
        }
    }

    // HTTP Handler
    public void service(HttpServletRequest request, HttpServletResponse response) 
    throws IOException
    {
        // public void handle(String target, Request request, HttpServletRequest _request,
        //                 HttpServletResponse response) throws IOException, ServletException {

        String method = request.getMethod();
        if(!method.equals("GET") && !method.equals("POST") && !method.equals("PUT") && !method.equals("DELETE") && 
                !method.equals("HEAD") && !method.equals("OPTIONS")) return;

        //request.setHandled(true);
        //parseGetParams(request, request.getRequestURI(), request.getQueryString());
        String identifier = request.getRequestURI();
        if(identifier.startsWith(STATIC_URI_PREFIX)) {
          handleStatic(request, response);
          return;
        }
        
        DOAuthentication auth = getAuthentication(request, response);
        System.err.println("got authentication: "+auth+" for request: "+request);
        if(auth==null) return;
        
        response.setStatus(HttpServletResponse.SC_OK);
        
        if(identifier.startsWith("/do")) handleRest(request,response, auth);
        else if(identifier.startsWith("/invoke")) handleInvoke(request,response);
        else handleWeb(request, response, auth);
    }


    void printLoginForm(HttpServletRequest req, HttpServletResponse response) {
        response.setStatus(HttpServletResponse.SC_OK);
        response.setContentType("text/html");
        // TODO: need to check cookie signature here!
        String loggedInAs = (String)req.getSession(true).getAttribute(OPENID_ATTRIBUTE);
        try {
            PrintWriter w = response.getWriter();
            w.print("<div id=\"openidauth\">");

            if(loggedInAs != null) {
                w.print("<span>You are logged in as: "+safeHTML(loggedInAs)+" - <a href=\"/auth/logout\">Logout</a></span>");
            }
            w.print("<form action=\"/auth/openid\" method=\"post\" id=\"openid_form\">");
            w.print("<input type=\"text\" size=\"30\" value=\"\" name=\"openid_url\" id=\"openid_url\"/>");
            w.print("<div> <input type=\"hidden\" name=\"signin\" value=\"true\"/>");
            w.print("<b>Login with your OpenID URL:</b>");
            w.print("<input type=\"submit\" value=\"Login\"/><br/>");
            w.print("<i>For example: <tt>http://someone.bloghost.com</tt></i>");
            w.print("</form> </div>");
            /*
       w.print("<br/><br/>");
       w.print("<img src=\"http://l.yimg.com/us.yimg.com/i/ydn/openid-signin-yellow.png\" alt=\"Sign in with Yahoo\" onclick=\"submitForm('http://www.yahoo.com');\"/>");
       w.print("<br/><br/>");
       w.print("<img src=\"http://buttons.googlesyndication.com/fusion/add.gif\" alt=\"Sign in with Google\" onclick=\"submitForm('https://www.google.com/accounts/o8/id');\"/>");
             */
            w.close();
        } catch (IOException e) {
            System.err.println("Error showing login form: "+e);
        }
    }


    private Map/*<String, String>*/ convertToStringValueMap(Map/*<String, String[]>*/ parameterMap) {
        Map/*<String,String>*/ ret = new HashMap();
        Set set = parameterMap.entrySet();
        for (Iterator iter = set.iterator(); iter.hasNext();) {
            Map.Entry mapEntry = (Map.Entry) iter.next();
            String key = (String) mapEntry.getKey();
            String[] value = (String[]) mapEntry.getValue();
            ret.put(key, value[0]);
        }
        return ret;
    }


    void handleInvoke(HttpServletRequest req, HttpServletResponse response) {
        // TODO
    }

    /** 
     * Handle a request (or update) for a DO component as a RESTful resource.  The path to 
     * the resource dictates the object ID and element or attribute name.

   The paths are similar to the following:
   <pre>
   /do/objectid/
   /do/objectid/att
   /do/objectid/el
   /do/objectid/el/elementid
   /do/objectid/el/elementid/att
   </pre>

     */
    private void handleRest(HttpServletRequest request, HttpServletResponse response, DOAuthentication auth) throws IOException {
        try {
            handleRestWrapped(request,response,auth);
        }
        catch(DOException e) {
            if(e.getErrorCode()==DOException.PERMISSION_DENIED_ERROR) {
                response.setHeader("WWW-Authenticate", "Basic realm=\"hdl:" + serverMain.getServerID() + "\"");
                response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
            }
        }
    }
    
    private void handleRestWrapped(HttpServletRequest request, HttpServletResponse response, DOAuthentication auth) throws IOException {
        response.setStatus(HttpServletResponse.SC_OK);

        String method = request.getMethod();
        boolean isGet = method.equals("GET") || method.equals("HEAD");
        boolean isPut = method.equals("PUT");
        boolean isPost = method.equals("POST");
        boolean isDelete = method.equals("DELETE");

        String pathStr = request.getRequestURI();
        if(!pathStr.startsWith("/do")) {
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }
        pathStr = pathStr.substring("/do".length());
        if(pathStr.startsWith("/")) {
            pathStr = pathStr.substring(1);
        }
        else if(!pathStr.isEmpty()) {
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }
        if(pathStr.endsWith("/")) {
            pathStr = pathStr.substring(0,pathStr.length()-1);
        }

        Repository repo = new HTTPRepositoryInterface(request, auth.getID());
        StringTokenizer pathTokens = new StringTokenizer(pathStr, "/", false);
        
        String objectID = null;
        DigitalObject digObj = null;
        
        // Handle /do alone
        if(!pathTokens.hasMoreTokens()) {
            if(isGet) {
                // Return XML Document Listing all DOs accessible through this DOR; if non, return empty list
                response.setContentType(XML);
                response.setCharacterEncoding("UTF-8");
                Iterator allObj = repo.listObjects();
                Writer w = response.getWriter();
                w.write("<dor id=\"");
                w.write(StringUtils.cgiEscape(repo.getID()));
                w.write("\">\n");
                while(allObj.hasNext()) {
                    String objID = String.valueOf(allObj.next());
                    w.write("    <do id=\"");
                    w.write(StringUtils.cgiEscape(objID));
                    w.write("\" />\n");
                }
                w.write("</dor>\n");
                return;
            } 
            else if(isPost) {
                if(request.getInputStream()!=null && request.getInputStream().read()!=-1) {
                    response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                    return;
                }
                
                digObj = repo.createDigitalObject(null);
                response.setStatus(HttpServletResponse.SC_CREATED);
                objectID = digObj.getID();
                response.setHeader("Location","/do/" + StringUtils.encodeURLComponent(objectID));
            }
            else {
                response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
                response.setHeader("Allow","GET, HEAD, POST");
                return;
            }
        }
        
        String token = null;
        if(pathTokens.hasMoreTokens()) token = urlDecode(pathTokens.nextToken());
        
        if(objectID==null) objectID = token;
        
        if(!pathTokens.hasMoreTokens()) {
            // either a POST to /do, or is /do/<id>
            if(isDelete) {
                if(repo.verifyDigitalObject(objectID)) {
                    repo.deleteDigitalObject(objectID);
                    response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                    return;
                }
                else {
                    response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                    return;
                }
            }
            else if((isPut || isPost) && token!=null) {
                /* Creates a new DO with the speciﬁed handle. If created, the DOR will assign
                 * assign the speciﬁed handle to the DO and return it in a response with 
                 * HTTP 201 Created. 
                 */
                if(request.getInputStream()!=null && request.getInputStream().read()!=-1) {
                    response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                    return;
                }

                if(!repo.verifyDigitalObject(objectID)) {
                    digObj = repo.createDigitalObject(objectID);
                    response.setStatus(HttpServletResponse.SC_CREATED);
                    String uri = "/do/" + pathStr;
                    response.setHeader("Location",uri);
                }
                else {
                    response.setStatus(HttpServletResponse.SC_CONFLICT);
                    return;
                }
            }
        }
        
        if(digObj==null) {
            if(!repo.verifyDigitalObject(objectID)) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                return;
            }
            digObj = repo.getDigitalObject(objectID);
        }

        if(!pathTokens.hasMoreTokens()) {
            response.setContentType(XML);
            response.setCharacterEncoding("UTF-8");
            Writer w = response.getWriter();
            w.write("<dor id=\"");
            w.write(StringUtils.cgiEscape(repo.getID()));
            w.write("\">\n");
            w.write("    <do id=\"");
            w.write(StringUtils.cgiEscape(objectID));
            w.write("\" >\n");
            HeaderSet atts = digObj.getAttributes();
            for(Iterator it=atts.iterator(); it.hasNext(); ) {
                HeaderItem attItem = (HeaderItem)it.next();
                w.write("        <att name=\"");
                w.write(StringUtils.cgiEscape(attItem.getName()));
                w.write("\" value=\"");
                w.write(StringUtils.cgiEscape(attItem.getValue()));
                w.write("\" />\n");
            }

            String elementIDs[] = digObj.listDataElements();
            for(int i=0; elementIDs!=null && i<elementIDs.length; i++) {
                DataElement element = digObj.getDataElement(elementIDs[i]);
                atts = element.getAttributes();
                w.write("        <el id=\"");
                w.write(StringUtils.cgiEscape(elementIDs[i]));
                w.write("\">\n");
                for(Iterator it=atts.iterator(); it.hasNext(); ) {
                    HeaderItem attItem = (HeaderItem)it.next();
                    w.write("            <att name=\"");
                    w.write(StringUtils.cgiEscape(attItem.getName()));
                    w.write("\" value=\"");
                    w.write(StringUtils.cgiEscape(attItem.getValue()));
                    w.write("\" />\n");
                }
                w.write("        </el>\n");
            }
            w.write("    </do>\n");
            w.write("</dor>\n");
            return;
        }

        token = pathTokens.nextToken(); 
        
        if(token.equals("att")) {
            String att = pathTokens.hasMoreTokens() ? urlDecode(pathTokens.nextToken()) : null;

            if(pathTokens.hasMoreTokens()) {
                response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                return;
            }
            
            if(att==null) {
                if(isGet) {
                    HeaderSet atts = digObj.getAttributes();
                    response.setContentType(XML);
                    response.setCharacterEncoding("UTF-8");
                    Writer w = response.getWriter();
                    w.write("<dor id=\"");
                    w.write(StringUtils.cgiEscape(repo.getID()));
                    w.write("\">\n");
                    w.write("    <do id=\"");
                    w.write(StringUtils.cgiEscape(objectID));
                    w.write("\" >\n");
                    for(Iterator it=atts.iterator(); it.hasNext(); ) {
                        HeaderItem attItem = (HeaderItem)it.next();
                        w.write("        <att name=\"");
                        w.write(StringUtils.cgiEscape(attItem.getName()));
                        w.write("\" value=\"");
                        w.write(StringUtils.cgiEscape(attItem.getValue()));
                        w.write("\" />\n");
                    }
                    w.write("    </do>\n");
                    w.write("</dor>\n");
                    return;
                }
                else {
                    response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
                    response.setHeader("Allow","GET, HEAD");
                    return;
                }
            }
            else {
                if(isGet) {
                    String value = digObj.getAttribute(att, null);
                    if(value==null) {
                        response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                        return;
                    }
                    response.setContentType(PLAINTEXT);
                    // we were given a specific attribute, so return it directly
                    response.getWriter().write(value);
                    return;
                }
                else if(isDelete) {
                    String value = digObj.getAttribute(att, null);
                    if(value==null) {
                        response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                        return;
                    }
                    digObj.deleteAttribute(att);
                    response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                    return;
                }
                else if (isPut || isPost) {
                    if(request.getHeader("Content-Encoding")!=null || request.getHeader("Content-Range")!=null) {
                        response.setStatus(HttpServletResponse.SC_NOT_IMPLEMENTED);
                        return;
                    }
                    
                    String value = streamToString(request.getInputStream(),request.getCharacterEncoding());
                    boolean attributeExisted = digObj.getAttribute(att, null)!=null;
                    // Replaces the plaintext value of this attribute. If created,returns HTTP 201 Created. 
                    digObj.setAttribute(att, value);

                    if(!attributeExisted) {
                        response.setStatus(HttpServletResponse.SC_CREATED);
                        response.setHeader("Location","/do/" + pathStr);
                    }
                    response.setContentType(PLAINTEXT);
                    // we were given a specific attribute, so return it directly
                    response.getWriter().write(value);
                    return;
                } 
                else {
                    response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
                    response.setHeader("Allow","GET, HEAD, POST, PUT, DELETE");
                    return;
                }
            }
        } else if(token.equals("el")) {
            // we are to return element information
            String elementID = pathTokens.hasMoreTokens() ? pathTokens.nextToken() : null;
            if(elementID==null) {
                // no element ID...  provide a list of elements
                if(isGet) {
                    response.setContentType(XML);
                    response.setCharacterEncoding("UTF-8");
                    Writer w = response.getWriter();
                    w.write("<dor id=\"");
                    w.write(StringUtils.cgiEscape(repo.getID()));
                    w.write("\">\n");
                    w.write("    <do id=\"");
                    w.write(StringUtils.cgiEscape(objectID));
                    w.write("\" >\n");
                    String elementIDs[] = digObj.listDataElements();
                    for(int i=0; elementIDs!=null && i<elementIDs.length; i++) {
                        DataElement element = digObj.getDataElement(elementIDs[i]);
                        w.write("        <el id=\"");
                        w.write(StringUtils.cgiEscape(elementIDs[i]));
                        w.write("\"/>\n");
                    }
                    w.write("    </do>\n");
                    w.write("</dor>\n");
                    return;
                } else {
                    response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
                    response.setHeader("Allow", "GET, HEAD");
                    return;
                }
            } else {
                // an element ID was provided, so we're looking for data or attributes
                boolean exists = digObj.verifyDataElement(elementID);
                DataElement element = digObj.getDataElement(elementID);
                String nextToken = pathTokens.hasMoreTokens() ? pathTokens.nextToken() : null;
                
                if(nextToken==null) {
                    if(isDelete) {
                        if(!exists) {
                            response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                            return;
                        }
                        digObj.deleteDataElement(elementID);
                        response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                        return;
                    }
                    else if(isGet) {
                        if(!exists) {
                            response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                            return;
                        }
                        String mimetype = element.getAttribute(DOConstants.MIME_TYPE_ATTRIBUTE,null);
                        if(mimetype!=null) response.setContentType(mimetype);
                        String filename = element.getAttribute(DOConstants.FILE_NAME_ATTRIBUTE, elementID);
                        response.setHeader("Content-disposition", "attachment; filename=\"" + escapeFilename(filename) + "\"");
                        response.setHeader("Content-Length", String.valueOf(element.getSize()));
                        
                        InputStream in = element.read();
                        OutputStream out = response.getOutputStream();
                        byte buf[] = new byte[BUFFER_SIZE];
                        int r;
                        while((r=in.read(buf))>=0) out.write(buf, 0, r);
                        out.close();
                        return;
                    } else if(isPut || isPost) {
                        if(request.getHeader("Content-Encoding")!=null || request.getHeader("Content-Range")!=null) {
                            response.setStatus(HttpServletResponse.SC_NOT_IMPLEMENTED);
                            return;
                        }

                        InputStream in = request.getInputStream();
                        if(in!=null) {
                            element.write(in);
                            //String mimetype = request.getContentType();
                            //if(mimetype!=null) element.setAttribute(DOConstants.MIME_TYPE_ATTRIBUTE,mimetype);
                        }
                        if(!exists) {
                            response.setStatus(HttpServletResponse.SC_CREATED);
                            response.setHeader("Location","/do/" + pathStr);
                            return;
                        }
                        else {
                            response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                            return;
                        }
                    } else {
                        response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
                        response.setHeader("Allow","GET, HEAD, PUT, POST, DELETE");
                        return;
                    }
                }
                
                else if(nextToken.equals("att")) {
                    String att = pathTokens.hasMoreTokens() ? urlDecode(pathTokens.nextToken()) : null;
                    if(pathTokens.hasMoreTokens()) {
                        response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                        return;
                    }

                    if(att==null) {
                        if(isGet) {
                            if(!exists) {
                                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                                return;
                            }

                            response.setContentType(XML);
                            response.setCharacterEncoding("UTF-8");
                            Writer w = response.getWriter();
                            w.write("<dor id=\"");
                            w.write(StringUtils.cgiEscape(repo.getID()));
                            w.write("\">\n");
                            w.write("    <do id=\"");
                            w.write(StringUtils.cgiEscape(objectID));
                            w.write("\" >\n");
                            HeaderSet atts = element.getAttributes();
                            w.write("        <el id=\"");
                            w.write(StringUtils.cgiEscape(elementID));
                            w.write("\">\n");
                            for(Iterator it=atts.iterator(); it.hasNext(); ) {
                                HeaderItem attItem = (HeaderItem)it.next();
                                w.write("            <att name=\"");
                                w.write(StringUtils.cgiEscape(attItem.getName()));
                                w.write("\" value=\"");
                                w.write(StringUtils.cgiEscape(attItem.getValue()));
                                w.write("\" />\n");
                            }
                            w.write("        </el>\n");
                            w.write("    </do>\n");
                            w.write("</dor>\n");
                            return;
                        } else {
                            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
                            response.setHeader("Allow", "GET, HEAD");
                            return;
                        }

                    }
                    else {
                        if(isGet) {
                            if(!exists) {
                                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                                return;
                            }
                            String value = element.getAttribute(att, null);
                            if(value==null) {
                                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                                return;
                            }
                            response.setContentType(PLAINTEXT);
                            // we were given a specific attribute, so return it directly
                            response.getWriter().write(value);
                            return;
                        }
                        else if(isDelete) {
                            if(!exists) {
                                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                                return;
                            }
                            String value = element.getAttribute(att, null);
                            if(value==null) {
                                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                                return;
                            }
                            element.deleteAttribute(att);
                            response.setStatus(HttpServletResponse.SC_NO_CONTENT);
                            return;
                        }
                        else if (isPut || isPost) {
                            if(request.getHeader("Content-Encoding")!=null || request.getHeader("Content-Range")!=null) {
                                response.setStatus(HttpServletResponse.SC_NOT_IMPLEMENTED);
                                return;
                            }

                            if(!exists) {
                                element.write(new ByteArrayInputStream(new byte[0]));
                            }

                            String value = streamToString(request.getInputStream(),request.getCharacterEncoding());
                            boolean attributeExisted = element.getAttribute(att, null)!=null;
                            // Replaces the plaintext value of this attribute. If created,returns HTTP 201 Created. 
                            element.setAttribute(att, value);

                            if(!attributeExisted) {
                                response.setStatus(HttpServletResponse.SC_CREATED);
                                response.setHeader("Location","/do/" + pathStr);
                            }
                            response.setContentType(PLAINTEXT);
                            // we were given a specific attribute, so return it directly
                            response.getWriter().write(value);
                            return;
                        } 
                        else {
                            response.setStatus(HttpServletResponse.SC_METHOD_NOT_ALLOWED);
                            response.setHeader("Allow","GET, HEAD, POST, PUT, DELETE");
                            return;
                        }
                    }
                } else {
                    response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
                    return;
                }
            }
        } else {
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
            return;
        }
    }

    /**
     * Generates a login form to allow users to log into the system with an OpenID
     * 
     * 
     * @param req
     * @param response
     * @return
     */
    String generateLoginForm(HttpServletRequest req, HttpServletResponse response) {
        response.setStatus(HttpServletResponse.SC_OK);
        response.setContentType("text/html");
        // TODO: need to check cookie signature here!
        String loggedInAs = (String) req.getSession(true).getAttribute(OPENID_ATTRIBUTE);

        StringBuffer s = new StringBuffer();
        s.append("<div id=\"openidauth\">");

        if (loggedInAs != null) {
            s.append("<span>You are logged in as: " + safeHTML(loggedInAs) + " - <a href=\"/auth/logout\">Logout</a></span></div>");
        } else {
            s.append("<form action=\"/auth/openid\" method=\"post\" id=\"openid_form\" align=\"right\">");
            s.append("<b>Maybe Login with your OpenID URL: </b>");
            s.append("<input type=\"text\" size=\"30\" value=\"\" name=\"openid_url\" id=\"openid_url\"/>");
            s.append("<input type=\"hidden\" name=\"signin\" value=\"true\"/>");
            s.append("<input type=\"submit\" value=\"Login\"/><br/>");
            s.append("<i>For example: <tt>http://someone.bloghost.com</tt></i>");
            s.append("</form> </div>");
        }
        return s.toString();
    }
    
    private Template getTemplate(String template) throws Exception {
      try {
        Template t = Velocity.getTemplate(template);
        if(t!=null) return t;
      } catch (Exception e) {}
      
      return Velocity.getTemplate("/net/cnri/apps/doserver/resources/pages/" + template);
    }
    
    
    private void handleStatic(HttpServletRequest request, HttpServletResponse response) 
    throws IOException
    {
      String path = request.getPathInfo();
      if(path.startsWith(STATIC_URI_PREFIX)) path = path.substring(STATIC_URI_PREFIX.length());
      
      File f = new File(staticBase, path);
      sendFile(request, response, f);
    }

    private void sendFile(HttpServletRequest request, HttpServletResponse response, File f)
    throws IOException
    {
      if(!f.exists() || !f.canRead()) {
        response.sendError(HttpServletResponse.SC_NOT_FOUND);
        return;
      }
      
      // TODO: bring http-range support over from the EMRUI retrieve controller
      
      FileInputStream fin = new FileInputStream(f);
      ServletOutputStream sout = response.getOutputStream();
      byte buf[] = new byte[BUFFER_SIZE];
      int r;
      while((r=fin.read(buf))>=0) sout.write(buf, 0, r);
      sout.close();
    }
    
    
    /**
     * Handle HTTP communication with a client (likely a web browser) to display dynamically rendered HTML.
     * Accepts browser requests for '/', '/show', '/list', '/search' and '/get'.
     * '/auth' requests are handled by the authentication method.
     * 
     * Powered by Velocity. Uses template built into the classpath.
     * 
     * @param request
     * @param response
     * @throws IOException
     */
    private void handleWeb(HttpServletRequest request, HttpServletResponse response, DOAuthentication auth) throws IOException {
        response.setContentType(HTML);
        Repository repo = new HTTPRepositoryInterface(request, auth.getID());
        SearchableMap model = new SearchableMap(repo);
        String repoName = serverMain.getServerID();
        String repoLink = objectHref(serverMain.getServerID());
        model.put("loginForm", generateLoginForm(request, response));
        model.put("repoName", repoName);
        model.put("repoLink", repoLink);
        model.put("baseurl", request.getRequestURL());
        //model.put("baseurl3", new java.net.URL(new java.net.URL(request.getRequestURI()), "/"));
        model.put("repo", new SearchableMap(repo));
        
        String templateName = "index.html";
        String requestURI = request.getRequestURI();
        
        if (requestURI.equals("/")) {
          // display the default page
          String objectID = request.getParameter(OBJECT_ID_PARAM);
          if(objectID!=null) model.put("object", model.getObject(objectID)); // add a specific object, if given
        } else if(requestURI.startsWith("/auth/basic")) {
          String redirectURL = request.getParameter("redirect");
          if(redirectURL==null || redirectURL.trim().length()<=0) redirectURL = "/";
          System.err.println("authenticated as "+auth+"; redirecting to: "+redirectURL);
          response.sendRedirect(redirectURL);
          return;
          
//        } else if (request.getRequestURI().equals("/list")) {
//          // TODO: get rid of the "big list".   all listings should be searches
//          listObjects(repo, request, response);
//          return;
        } else if (requestURI.equals("/show")) {
          String objectID = request.getParameter(OBJECT_ID_PARAM);
          if(objectID==null) objectID = repo.getID();
          model.put("object", model.getObject(objectID));
          templateName = "showObject.html";
        } else if (requestURI.equals("/search")) {
          String queryStr = request.getParameter("q");
          if(queryStr==null) queryStr = "";
          model.put("query", queryStr);
          model.put("results", model.doSearch(queryStr));
          templateName = "searchResults.html";
//        } else if(requestURI.equals("/create")) {
//          handleCreate(request, response, auth);
//          return;
        } else if (requestURI.equals("/get")) { // an inline request for an element
          String objectID = request.getParameter("id");
          try {
            DigitalObject digObj = repo.getDigitalObject(objectID);
            String elementID = request.getParameter("el");
            if(elementID==null) {
              elementID = digObj.getAttribute(DEFAULT_ELEMENT_REF_ATT, DEFAULT_GET_ELEMENT);
            }
            DataElement element = digObj.getDataElement(elementID);
            String filename = element.getAttribute(DOConstants.FILE_NAME_ATTRIBUTE, elementID);
            response.setContentType(element.getAttribute(DOConstants.MIME_TYPE_ATTRIBUTE, DEFAULT_MIME_TYPE));
            response.setHeader(CONTENT_DISPOSITION_HEADER, "inline; filename=\"" + safeHTML(filename) + "\";");
            
            long size = element.getSize();
            if (size >= 0) response.setContentLength((int) size);
            InputStream in = element.read();
            OutputStream out = response.getOutputStream();
            byte buf[] = new byte[BUFFER_SIZE];
            int r;
            while ((r = in.read(buf)) >= 0) out.write(buf, 0, r);
          } catch (DOException e) {
            System.err.println("Error accessing object: " + e);
            if(!response.isCommitted()) {
              response.setStatus(HttpServletResponse.SC_NOT_FOUND);
              PrintWriter w = response.getWriter();
              w.println("<html>");
              w.println("<head><title>Object not found: " + safeHTML(objectID) + "</title></head>");
              w.println("<body>");
              w.println("<h1>Object not found: " + safeHTML(objectID) + "</h1>");
            }
          }
          return;
        } else {
          String resourcePath = request.getPathInfo();
          System.err.println("Unrecognized resource path: "+resourcePath);
          response.setStatus(HttpServletResponse.SC_NOT_FOUND);
          response.getWriter().println("<html>");
          response.getWriter().println("<head><title>Requested resource not available</title></head>");
          response.getWriter().println("<body><p>You are trying to access a resource that is unavailable. Its likely that the URL is incorrect.</p></body></html>");
          return;
        }
        
        String overrideTemplate = request.getParameter("template");
        if(overrideTemplate!=null) {
          templateName = overrideTemplate;
        }
        
        try {
          getTemplate(templateName).merge(new VelocityContext(model), response.getWriter());
        } catch (Exception e) {
          e.printStackTrace();
        }
    }

    
//    /**
//     * process a request to create an object, specifying the attributes as HTTP parameters
//     * @param request
//     * @param response
//     * @param auth
//     * @throws IOException
//     */
//    private void handleCreate(HttpServletRequest request, HttpServletResponse response, DOAuthentication auth) throws IOException {
//      // TODO: put generic object-factory here that creates an object and populates it with attributes and elements
//      Map<String,String> queryParams = MetadataUtil.parseQueryParameters(request.getQueryString());
//      ...;
//      DigitalObject dobj = repo.createDigitalObject(null);
//      DataElement content = dobj.getDataElement(DOConstants.CONTENT_ELEMENT_ID);
//      content.write(req.getInputStream());
//      for(Map.Entry<String,String> item : queryParams.entrySet()) {
//          if(item.getKey().equals(DOConstants.MIME_TYPE_ATTRIBUTE) || item.getKey().equals(DOConstants.FILE_NAME_ATTRIBUTE)) {
//              content.setAttribute(item.getKey(),item.getValue());
//          }
//          else {
//              dobj.setAttribute(item.getKey(),item.getValue());
//          }
//      }
//      // MainController.newObjects.add(dobj);
//      resp.getWriter().println(dobj.getID());
//      resp.getWriter().close();
//      return null;
//
//      
//      
//      
//    }
    
    // Two private classes encapsulate data for Velocity access
    public static class AttributePair {
        String key;
        String val;

        public AttributePair(String a, String b) {
            key = a;
            val = b;
        }

        public String getKey(){
            return key;
        }

        public String getVal(){
            return val;
        }
    }

    public static class ElementType {
        String elementLink;
        ArrayList<AttributePair> attributes = new ArrayList<AttributePair>();

        public ElementType(String link) {
            elementLink = link;
        }

        public String getElementLink(){
            return elementLink;
        }

        public ArrayList<AttributePair> getAttributes(){
            return attributes;
        }
    }
    
    
//    private void listObjects(Repository repo, HttpServletRequest request, HttpServletResponse response) throws IOException {
//        response.reset();
//        response.setContentType(HTML);
//        response.setStatus(HttpServletResponse.SC_OK);
//
//        String login_form = generateLoginForm(request, response);
//
//        ArrayList<String> objectLinks = new ArrayList<String>();
//        for (Iterator it = repo.listObjects(); it.hasNext();)
//            objectLinks.add(objectHref(String.valueOf(it.next())));
//
//        VelocityContext context = new VelocityContext();
//        context.put("repoID", safeHTML(repo.getID()));
//        context.put("objectLinks", objectLinks);
//        context.put("loginForm", login_form);
//
//        try {
//            Template template = getTemplate("browseObjects.html");
//            template.merge(context,response.getWriter());
//        } catch (Exception e) {
//            System.err.println("Problem merging template : " + e);
//        }
//
//    }

    private boolean checkAuthentication(DOAuthentication auth) {
      try {
        return DOConnection.getResolver().checkAuthentication(auth.toHandleAuth());
      }
      catch (Exception e) {
        return false;
      }
    }
    
    /** Get the client's authentication, if any.  If the user has not authenticated 
     * this will return an anonymous auth object.  If the user is being authenticated
     * then this will provide an authentication method to the client (usually a 
     * basic http not-authorized response) and return null.  Callers should interpret
     * the null response as an indication that the request should not continue. */
    private DOAuthentication getAuthentication(HttpServletRequest request, HttpServletResponse response) {
      DOAuthentication auth = (DOAuthentication)request.getAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE);
      HttpSession session = request.getSession(true);
      String sessionAuthHeader = (String)session.getAttribute(AUTHORIZATION_ATTRIBUTE);
      if(auth==null) {
        auth = (DOAuthentication)session.getAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE);
      }
      
      String method = request.getMethod();
      String path = request.getRequestURI();
      String baseURL = request.getScheme()+"://"+request.getServerName()+":"+request.getServerPort();
      
      String authHeader = request.getHeader("Authorization");
      if(method.equals("GET") && (path.equals("/auth/logout") || path.equals("/logout/"))) {
        auth = null;
        request.removeAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE);
        session.removeAttribute(OPENID_ATTRIBUTE);
        session.removeAttribute(AUTHORIZATION_ATTRIBUTE);
        session.removeAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE);
        //session.invalidate();
        try {
          //FIXME:  This can't be the best redirect destination
          response.sendRedirect(request.getContextPath());
        } catch (IOException e) {
          System.err.println("Error sending redirect during logout: "+e);
        }
        return null;
      }
      else if(method.equals("POST") && path.equals("/auth/openid")) {
        // the user is requesting to login using an OpenID URL
        //String context = request.getContextPath();
        //if(context!=null) baseURL += context;
        // http://www.myopenid.com/server?
        // openid.trust_root=http%3A%2F%2Flocalhost%3A8080%2Fjoid&
        // openid.ns=http%3A%2F%2Fspecs.openid.net%2Fauth%2F2.0&
        // openid.identity=http%3A%2F%2Fseanreilly.com&openid.claimed_id=http%3A%2F%2Fseanreilly.com&
          // openid.mode=checkid_setup&
          // openid.realm=http%3A%2F%2Flocalhost%3A8080%2Fjoid&
          // openid.assoc_handle=%7BHMAC-SHA1%7D%7B4aa015d7%7D%7BP3eXSg%3D%3D%7D&
          // openid.return_to=http%3A%2F%2Flocalhost%3A8080%2Fjoid
          try {
              String id = request.getParameter("openid_url");
              if(id==null || id.trim().length()<=0) {
                  response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Missing OpenID URL parameter (openid_url)");
                  return null;
              }
              String trustRoot = baseURL.toString();
              String s = joid.getAuthUrl(id, baseURL, baseURL);
              response.sendRedirect(s);
              auth = null;
              request.removeAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE);
              session.removeAttribute(OPENID_ATTRIBUTE);
              session.removeAttribute(AUTHORIZATION_ATTRIBUTE);
              session.removeAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE);
              return null; // the user will bounce back and at that time the authentication will be setup
          } catch (Exception e) {
              e.printStackTrace(System.err);
          }
      } else if(method.equals("GET") && (path.equals("/auth/login") || path.equals("/login/"))) {
          printLoginForm(request, response);
          return null;
      } else if(authHeader!=null && !authHeader.equals(sessionAuthHeader)) {
          // if authentication provided in the http headers, use it
          System.err.println("authentication header: "+authHeader);
          if(StringUtils.fieldIndex(authHeader, ' ', 0).equalsIgnoreCase("Basic")) {
              String userpass = Util.decodeString(base64Decode(StringUtils.fieldIndex(authHeader, ' ', 1)));
              auth = new SecretKeyAuthentication(StringUtils.fieldIndex(userpass, ':', 0),
                      Util.encodeString(StringUtils.fieldIndex(userpass, ':', 1)));

              boolean verifiedAuth = false;
              try {
                  verifiedAuth = DOClient.getResolver().checkAuthentication(auth.toHandleAuth());
              } catch (Exception e) {
                  System.err.println("Error verifying authentication: "+e);
                  verifiedAuth = false;
              }

              if(verifiedAuth) {
                  System.err.println("authentication verified: "+auth);
                  request.setAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE, auth);
                  session.setAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE, auth);
                  session.setAttribute(AUTHORIZATION_ATTRIBUTE,authHeader);
                  session.removeAttribute(OPENID_ATTRIBUTE);
                  return auth;
              } else {
                  // unable to verify authentication... try again
                  response.setHeader("WWW-Authenticate", "Basic realm=\"hdl:" + serverMain.getServerID() + "\"");
                  response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
                  try {
                      response.getWriter().write("Authentication failed");
                      response.getWriter().close();
                  } catch (Exception e2) {}
                  return null;
              }
          }
      } else if(path.equals("/auth/basic")) {
          response.setHeader("WWW-Authenticate", "Basic realm=\"hdl:" + serverMain.getServerID() + "\"");
          response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
          try {
              response.getWriter().write("Authentication failed");
              response.getWriter().close();
          } catch (Exception e) {}
          return null;
      }
      else if(request.getMethod().equals("GET") && request.getParameter(OPENID_IDENTITY) != null) {
          // the user is attempting to establish authentication using OpenID
          try {
              AuthenticationResult result = joid.authenticate(convertToStringValueMap(request.getParameterMap()));
              String identity = result.getIdentity();
              if(identity!=null) {
                  Cookie cookie = new Cookie(OPENID_IDENTITY, identity);
                  //if(cookieDomain != null) cookie.setDomain(cookieDomain);
                  //if(cookieMaxAge != null) cookie.setMaxAge(cookieMaxAge.intValue());
                  response.addCookie(cookie);

                  auth = new OpenIDDOAuthentication(identity);
                  request.setAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE, auth);
                  session.setAttribute(OpenIdFilter.OPENID_ATTRIBUTE, identity);
                  session.setAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE, auth);

                  // redirect to get rid of the long url
                  response.sendRedirect(result.getResponse().getReturnTo());
                  return null;
              }
          } catch(Exception e){
              e.printStackTrace();
              System.err.println("OpenID authentication failed: " + e.getMessage());
              return null;
          }
      }

      if(auth==null) {
        auth = PKAuthentication.getAnonymousAuth();
        request.setAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE, auth);
        session.setAttribute(DO_AUTHENTICATION_REQ_ATTRIBUTE, auth);
      }
      return auth;
    }
    
    private class OpenIDDOAuthentication implements DOAuthentication {
      private String openID;
      
      OpenIDDOAuthentication(String openID) {
        this.openID = openID;
      }
      
      public String getID() { return openID; }
      
      public void signChallenge(HeaderSet challenge, HeaderSet response)
      throws Exception 
      {
        throw new DOException(DOException.SERVER_ERROR, "OpenID authentication placeholder was asked to sign a challenge, which is not possible because the user keeps the credentials.");
      }
      
      /** Returns any certificates that supplement this entities authentication */
      public Certificate[] getCredentials() { return PKAuthentication.getAnonymousAuth().getCredentials(); }
      
      /** 
       * Returns this authentication in a form that will work with handle system
       * administration. */
      public net.handle.hdllib.AuthenticationInfo toHandleAuth() {
        return PKAuthentication.getAnonymousAuth().toHandleAuth();
      }
    }
    
    
    //Read from stream and return string
    private static String streamToString(InputStream input, String encoding) throws IOException{
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        byte buf[] = new byte[BUFFER_SIZE];
        int r;
        while((r=input.read(buf))>=0)
            bout.write(buf, 0, r);
        if(encoding==null) return Util.decodeString(bout.toByteArray());
        else {
            return new String(bout.toByteArray(), encoding);
        }
    }

    public static final String safeHTML(String str) {
        return StringUtils.cgiEscape(str);
    }

    private static final String urlDecode(String s) {
        return StringUtils.decodeURLIgnorePlus(s);
    }

    private static final String objectHref(String objID) {
        String href = "/show?id=" + StringUtils.encodeURLComponent(objID);
        return "<a class=\"objref\" href=\"" + safeHTML(href) + "\" >"+safeHTML(objID)+"</a>";
    }

    private static final String elementHref(String objID, String elementID) {
        String href = "/get?id="+StringUtils.encodeURLComponent(objID) + (elementID==null ? "" : ("&el="+StringUtils.encodeURLComponent(elementID)));
        return "<a class=\"getref\" href=\"" + StringUtils.cgiEscape(href) +
        "\" >"+safeHTML(elementID)+"</a>";
    }

    private static final byte[] base64Decode(String s) {
        ArrayList<Byte> res = new ArrayList<Byte>(s.length()*3/4);
        int len = s.length();
        int pos = 0;
        int i = 0; 
        int acc = -1;
        while(true) {
            int b = -1;
            while (b<0) {
                if(i>=len) break;
                char ch = s.charAt(i);
                i++;
                if(ch>='A' && ch<='Z') b = ch - 'A';
                else if(ch>='a' && ch<='z') b = ch - 'a' + 26;
                else if(ch>='0' && ch<='9') b = ch - '0' + 52;
                else if(ch=='+') b = 62;
                else if(ch=='/') b = 63;
                else if(ch=='=') break;
            }
            if(b<0) {
                // if(acc>=0) res.add((byte)acc);
                break;
            }
            if(pos==0) {
                acc = b << 2;
                pos = 1;
            }
            else if(pos==1) {
                acc |= b >>> 4;
                res.add((byte)acc);
                acc = (b & 0xF) << 4;
                pos = 2;
            }
            else if(pos==2) {
                acc |= b >>> 2;
                res.add((byte)acc);
                acc = (b & 0x3) << 6;
                pos = 3;
            }
            else {
                acc |= b;
                res.add((byte)acc);
                acc = -1;
                pos = 0;
            }
        }
        int size = res.size();
        byte[] resArray = new byte[size];
        for(i = 0; i<size; i++) resArray[i] = res.get(i);
        return resArray;
    }


    private class HTTPRepositoryInterface
    extends InternalRepository
    {
      private HttpServletRequest request;
      private String callerID;
      
      private HTTPRepositoryInterface(HttpServletRequest request, String callerID) 
        throws DOException
      {
        super(serverMain);
        this.request = request;
        this.callerID = callerID;
      }
      
      
      protected DOServerOperationContext getOpContext(String objectID, String operationID, HeaderSet params) {
        return new HTTPOperationContext(request, callerID, objectID, operationID, params);
      }
    }
    
    
    private class HTTPOperationContext
    extends DOServerOperationContext
    
    
    {
        private HttpServletRequest httpRequest;
        private HttpSession httpSession;

        private String callerID;
        private String operationID;
        private String objectID;
        HeaderSet params;
        private HashMap sessionInfo;
        private StorageProxy storage;

        private HTTPOperationContext(HttpServletRequest httpRequest, String callerID,
                String objectID, String operationID, HeaderSet params) {
            this.httpRequest = httpRequest;
            this.httpSession = httpRequest.getSession(true);
            this.callerID = callerID;
            this.operationID = operationID;
            this.objectID = objectID;
            this.params = params==null ? new HeaderSet() : params;
        }


        public InetAddress getClientAddress() {
            try {
                return InetAddress.getByName(httpRequest.getRemoteAddr());
            } catch (Exception e) {
                return null;
            }
        }

        /** If the client authenticated with a "forwardable" authentication then this will be
         * used to generate a DOClient using that authentication.  For the time being, we will
         * only return an client with anonymous identity. */
        public DOClient getDOClientWithForwardedAuthentication(String clientID) {
            return anonymousDOClient;
        }

        // the caller's authentication is established before the operation is invoked
        public boolean authenticateCaller() { return true; }

        public String[] getCredentialIDs() { return null; } // no group credentials through HTTP (yet)

        public boolean authenticateCredential(String credentialID) { return false; } // no group credentials through HTTP (yet)

        public String getCallerID() { return callerID; }

        public String getOperationID() { return operationID; }

        public String getTargetObjectID() { return objectID; }

        public String getServerID() { return serverMain.getServerID(); }

        public HeaderSet getOperationHeaders() { return params; }

        public synchronized StorageProxy getStorage() {
          if(storage!=null) return storage;
          HeaderSet txnMetadata = new HeaderSet();
          txnMetadata.addHeader("callerid", callerID);
          txnMetadata.addHeader("operationid", operationID);
          txnMetadata.addHeader("params", params);
          storage = new ConcreteStorageProxy(serverMain.getStorage(),
                                             serverMain.getServerID(),
                                             objectID, txnMetadata);
          return storage;
        }


        public void setConnectionMapping(Object mappingKey, Object mappingData) {
            // NOTE: HTTP mappings use the string value of the mappingKey, so be careful!
            if(mappingData==null) {
                httpSession.removeAttribute(String.valueOf(mappingKey));
            } else {
                httpSession.setAttribute(String.valueOf(mappingKey), mappingData);
            }
        }


        public Object getConnectionMapping(Object mappingKey) {
            // NOTE: HTTP mappings use the string value of the mappingKey, so be careful!
            return httpSession.getAttribute(String.valueOf(mappingKey));
        }


        /**
         * Performs the specified operation with the identity of the caller, or as 
         * the container repository if the forwarding operations are configured to use
         * the repository's own identity for forwarded operations.
         * If the serverID is null then the DO client that performs this request
         * will resolve the objectID to find the server.
         */
        public void performOperation(String serverID, String objectID,
                String operationID, HeaderSet params,
                java.io.InputStream input, java.io.OutputStream output)
        throws DOException 
        {
            HTTPOperationContext derivedCtx = new HTTPOperationContext(httpRequest, callerID, 
                    objectID, operationID,
                    params);
            try {
                serverMain.performOperation(derivedCtx, input, output);
            } catch (Exception e) {
                e.printStackTrace(System.err);
                if(e instanceof DOException) throw (DOException)e;
                throw new DOException(DOException.SERVER_ERROR,
                        "Error invoking derived operation: "+e);
            }

        }

    }

    public String escapeFilename(String filename) {
      // This is gruesome, but as of mid-2010 is the only reasonably portable solution
      try {
        String res = new String(filename.getBytes("ISO-8859-1"),"ISO-8859-1");
        return res.replace("\n","?").replace("\\","\\\\").replace("\"","\\\"");
      } catch(UnsupportedEncodingException e) {
        throw new AssertionError(e);
      }
    }

}