/*************************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
\*************************************************************************/

package net.cnri.dobj;

import java.security.*;
import java.security.cert.*;
import javax.net.ssl.*;
import java.util.*;



/** This object is used to determine whether the other side of a secure socket 
  * connection is authenticated or not based on their Handle and Digital Object
  * authentication.
  */
public class DOSSLTrustManager 
  implements X509TrustManager
{
  private X509Certificate[] acceptedIssuers = {};
  private Hashtable cachedPubkeys = new Hashtable();
  
  public DOSSLTrustManager() {
  }
  
  public void checkClientTrusted(X509Certificate chain[], String authType) 
  throws CertificateException
  {
    verifyCertificate(chain);
    
  }
  public void checkServerTrusted(X509Certificate chain[], String authType) 
  throws CertificateException
  {
    verifyCertificate(chain);
  }
  
  
  private void verifyCertificate(X509Certificate chain[]) 
    throws CertificateException
  {
    if(chain==null || chain.length<=0) {
      throw new CertificateException("No client certificate provided");
    }
    
    int numCerts = 0;
    try {
      for(int i=0; i<chain.length; i++) {
        X509Certificate cert = chain[i];
        if(cert==null) throw new CertificateException("Broken certificate chain at index "+i);
        
        cert.checkValidity();
        
        verifyCertByResolving(cert);
        numCerts++;
      }
    } catch (Exception e) {
      if(e instanceof CertificateException) throw (CertificateException)e;
      throw new CertificateException("Error verifying certificate: "+e);
    }
    
    if(numCerts<=0) {
      throw new CertificateException("No certificates provided by peer");
    }
  }
  
  public X509Certificate[] getAcceptedIssuers() {
    return acceptedIssuers;
  }
  
  
  private void verifyCertByResolving(X509Certificate cert) 
    throws CertificateException
  {
    String subject = extractHandleFromPrincipal(cert.getSubjectDN().getName());
    String issuer = extractHandleFromPrincipal(cert.getIssuerDN().getName());
    
    PublicKey clientKeys[] = resolvePublicKeys(issuer);
    if(clientKeys==null || clientKeys.length<=0) {
      throw new CertificateException("No public keys to validate peer certificate ");
    }
    
    Exception verifyException = null;
    PublicKey clientKey = null;
    for(int i=0; i<clientKeys.length; i++) {
      PublicKey testKey = clientKeys[i];
      if(testKey==null) continue;
      try {
        cert.verify(testKey);
        clientKey = testKey;
      } catch (Exception e) {
        verifyException = e;
      }
    }
    
    // if none of the peer's keys validate the certificate then the whole chain is invalid
    if(clientKey==null) {
      if(verifyException!=null) {
        if(verifyException instanceof CertificateException) {
          throw (CertificateException)verifyException;
        } else {
          throw new CertificateException("Unable to verify credential for '"+issuer+"': "+
                                         verifyException);
        }
      } else {
        throw new CertificateException("Unable to verify credential for '"+issuer+"'");
      }
    }
  }
  
  private static final String extractHandleFromPrincipal(String principal) {
    int idx = 0;
    int len = principal.length();
    while(idx<len) {
      int eqIdx = principal.indexOf('=', idx);
      if(eqIdx<0) return null;
      String key = principal.substring(idx, eqIdx);
      idx = eqIdx+1;
      eqIdx = principal.indexOf(',', idx);
      String val = null;
      if(eqIdx<0) {
        val = principal.substring(idx, len);
        idx = len;
      } else {
        val = principal.substring(idx, eqIdx);
        idx = eqIdx+1;
      }
      if(key.equals("O")) {
        return val;
      }
    }
    return null;
  }
  
  
  /** Securely resolve a public key for the given handle */
  private synchronized PublicKey[] resolvePublicKeys(String clientID) {
    if(cachedPubkeys.containsKey(clientID)) {
      return (PublicKey[])cachedPubkeys.get(clientID);
    }
    try {
      PublicKey keys[] = DOConnection.getResolver().resolvePublicKeys(clientID);
      
      if(keys==null || keys.length<=0) {
        System.err.println("No VALID public key is associated with ID "+clientID);
        return null;
      }
      
      cachedPubkeys.put(clientID, keys);
      return keys;
    } catch (Exception e) {
      System.err.println("error looking up public key for "+clientID+": "+e);
    }
    return null;
  }
  
  
}


